//
//     WARNING
//       This code is completely open and available to be modified and changed
//       in any way and by anybody. It is yours once you start to change it.
//       Don't try to patent your derivative code and sue us for infringement.
//       We were here first!
//
//       The reason it is provided this way is for your education and to illustrate some coding
//       ideas. It can cause damage to your Arduino or other hardware if you are not extremely careful.
//       No guarantee is offered and no risk is assumed by the code author.
//       All the risk is on you even if you follow the commments provided in the
//       source code or elsewhere. This is powerful code but you must know it well.
//
//       This code implements the Command Module as imagined by Robot-DHQ. The main purpose of
//       the code is to organize and send commands to one or more Robot-DHQ modules.
//
//       Robot-DHQ is the Robot Design Headquarters
//
//       The human interface paradigm is "immediate response" to any input. No CR or LF is required if
//       the input is a key on a keyboard. Similarly a button press sends a command immediately and
//       the code does not wait for the button to be released. Electrically the button press could
//       be the low to high transition or the high to low transition. If the button is behaving
//       "wrongly" one option is to make a change to the code.
//
//       It has many undocumented features and has not been optimized. You can learn
//       a lot by studying it or you can use it according to the documentation and
//       not read any further. Notes within the source code will give you some ideas
//       for using some of the undocumented features.
//
//       Consider it a "work in progress". The Command Module is described in
//       this document: D:\Projects\Arduino\Smart Modules\??? Command Module.odt
//
//       This code is designed to run on an Arduino Nano installed on a Robot-DHQ adapter board.
//       The serial port on the Arduino Nano accepts commands and displays the command menus.
//       The port is configured to operate an 115,200 baud.
//       This is the first thing to check if the menu is not readable when first connected.
//
//       The commands sent to Robot-DHQ modules are sent over a second serial interface.
//       The Arduino Nano only has one hardware serial interface so the second serial
//       interface is a software serial port. As a consequence there are a number of limitations.
//       One limitation is the need to operate this port at 57,600 baud, max.
//       Also, some delays are required in code to make the output readable.
//       Communication between modules is at 57,600 baud.
//
// how to avoid destroying the Arduino Nano
// 1   the Arduino will not be detected over the USB interface if the Schottky diode
//     has been destroyed. This can happen the +5 volt supply is accidentally
//     shorted to ground. The diode can only handle about 0.5 amps before it blows out.
//     This is therefore the first thing to suspect if the serial connection drops.
//     Therefore, always check for shorts before applying power
//
// 2   input pins are relatively safe to implement as long as they are not
//     connected to a power source with a voltage greater or less than the
//     hardware limitations, usually VCC and GROUND. The S1025 command is
//     provided to produce an input only state.
//

#include <SoftwareSerial.h>

SoftwareSerial gtSerial(9, 11); // RX is Arduino pin D9, TX is Arduino pin D11

// these variables affect the configuration settings at startup
//   timing parameters
//     the timing defaults can be adjusted by changing these parameters here
const int slow_loop_size = 2;
// implementing a pseudo millisecond timer
const unsigned long fast_loop_size = 1000;
// end of configuration constants

//   input and output routing
//     the input from the two serial ports is processed the same
//     it is processed when the CR character (13) is received
//     the identical output is sent to both serial outputs
//     the onboard Arduino Nano LED is blinked to indicate ??

// timing variables
unsigned long startMicros = 0;
unsigned long currentMicros = 0;

// new variables
byte rx_byte = 0;        // stores received byte
int display_menu = 0;
int sequence_steps_to_do = 0;
int repeats_to_do = 0;
int repeating = 0;
// the default action is to immediately show all incoming data from the remote node
// this can be "skipped" when a sequence is collecting data for a report, etc.
int step_no = 0;
int menu_number = 0;
int menu_item = 0;
int cmd_menu = 0;
int seq_menu = 8;
int input_mode = 0;
int digit_key = 0;
// settable parms
int digital_pin = 2;
int analog_pin = 0;
int parm_value_V = 0;
int parm_value_W = 0;
int parm_value_X = 0;
int parm_value_Y = 0;
int parm_value_Z = 0;
int temp_value = 0;
int entered_value = 0;
int digit_count = 0;
int history_item = 0;
int h_idx = 0;
int data_pos = 0;
int requestFunction = 0;
int requestMenu = 0;
int requestLine = 0;
String myStr = "";
String tmpString;
const int NUMBER_OF_ELEMENTS = 16;
const int MAX_SIZE = 12;
char cmdHistory [NUMBER_OF_ELEMENTS] [MAX_SIZE];
char seqCommand [NUMBER_OF_ELEMENTS] [MAX_SIZE];
int seqDelay [NUMBER_OF_ELEMENTS];
char cmdReceived[81];
char savResponse[20];
int sourcePin = 0;
int pulseWidth = 0;
int parmOne = 0;
int parmTwo = 0;

// indexes and counters
int pin_idx;
int led_idx = 0;
int slow_counter = 0;
int seq_step = 0;
int p_seq_step = 0;
int menu_option = 0;

// bit bang vars
const int SHORT_PERIOD = 25;
const int LONG_PERIOD = 250;
const int START_MSG_PERIOD = 10000;
const int END_CHAR_PERIOD = 10000;
int tx_period_1;
int tx_period_2;
int tx_period_3;
int tx_period_4;
int tx_period_5;
int tx_period_6;
int tx_period_7;

void setup() {
  pinMode(13, OUTPUT);
  digitalWrite(13, HIGH);
  gtSerial.begin(57600);  // software serial port
  Serial.begin(57600);
  while (!Serial) {
    ; // wait for serial port to connect. Needed for native USB port only
  }

  initialize_arrays();
  Serial.print(F("TextSource Module"));
}

void initialize_arrays() {
  history_item = 0;
  for (h_idx = 0; h_idx < 14; h_idx++) {
    myStr = "";
    myStr.toCharArray(cmdHistory[h_idx], 12);
  }
}

void loop() {
  currentMicros = micros();
  if (currentMicros - startMicros >= fast_loop_size) {
    // check clock and send queued and scheduled commands to remote processor.
    slowLoop();          // once per millisecond, 1000 times per second, period = 1000 microseconds
    startMicros = currentMicros;
  }

  // each character is processed as a complete command as it is entered (CR not required)
  while (Serial.available() > 0) {
    rx_byte = Serial.read();
    processGtSerialData();
    if (display_menu > 0) {
      // the "console" includes menus, prompts and help displays
      consoleRefresh();
      display_menu = 0;
    }
  }

  // check if byte available on the software serial port
  if (gtSerial.available()) {
    // get the byte from the software serial port
    rx_byte = gtSerial.read();
    processGtSerialData();
  }
}

void slowLoop() { // millisecond
  slow_counter++;
  if (slow_counter > slow_loop_size) {
    slow_counter = 0;
    // these actions are done once per 250 millisecond, 4 times per second, period = 250,000 microseconds
    //blink_LED();
  }
  if (sequence_steps_to_do > 0) {
    Serial.println();
    Serial.println(F("Command sequence is complete."));
    sequence_steps_to_do = 0;
  }
}

void processGtSerialData() {
  if (rx_byte != 10) { // newline
    if (rx_byte > 31) {
      if (rx_byte < 127) {
        if (data_pos < 80) {
          cmdReceived[data_pos] = rx_byte;
          data_pos++;
        }
      }
    }
  }
  if (rx_byte == 13) {
    cmdReceived[data_pos] = '\0';
    if (data_pos > 2) {
      processRequest();
    }
    data_pos = 0;
  }
}

void processRequest() {
  Serial.println();
  Serial.print(F("Cmd : "));
  Serial.println(cmdReceived);
  processChar1();
  processChar2();
  processChar3();
  Serial.print(F("Fnc : "));
  Serial.print(requestFunction);
  Serial.print(F(", Mnu : "));
  Serial.print(menu_number);
  Serial.print(F(", Itm : "));
  Serial.println(menu_item);
  switch (requestFunction) {
    case 1:
      res_locator(menu_number, menu_item, 0); // res_type 0 to return the menu text
      break;
    case 2:
      res_locator(menu_number, menu_item, 1); // res_type 1 to return the command
      break;
    default:
      Serial.print(F("(TextSource) The first character of the request, "));
      Serial.print(cmdReceived[0]);
      Serial.println(F(", is not a valid function."));
      break;
  }
}

// char 1 is the function
void processChar1() {
  switch (cmdReceived[0]) {
    case 'S':
      requestFunction = 1;
      break;
    case 'M': // return the text for a menu line item
      requestFunction = 1;
      break;
    case 'C': // return the text of a command
      requestFunction = 2;
      break;
    default:
      requestFunction = cmdReceived[0];
      break;
  }
}

// char 2 is the first parameter (menu number)
void processChar2() {
  switch (cmdReceived[1]) {
    case 48:
    case 49:
    case 50:
    case 51:
    case 52:
    case 53:
    case 54:
    case 55:
    case 56:
    case 57:
      menu_number = cmdReceived[1] - 48;
      break;
    default:
      menu_number = cmdReceived[1];
      break;
  }
}

// char 3 is the second parameter (menu line)
void processChar3() {
  switch (cmdReceived[2]) {
    case 48:
    case 49:
    case 50:
    case 51:
    case 52:
    case 53:
    case 54:
    case 55:
    case 56:
    case 57:
      menu_item = cmdReceived[2] - 48;
      break;
    case 69: // E - menu option key
      menu_item = 10;
      break;
    case 70: // F - menu option key
      menu_item = 11;
      break;
    case 71: // G - menu option key
      menu_item = 12;
      break;
    case 72: // H - menu option key
      menu_item = 13;
      break;
    default:
      menu_item = cmdReceived[2];
      break;
  }
}

void numericResponse() {
  entered_value = 0;
  digit_count = 0;
  while (digit_count < 5) {
    entered_value = entered_value * 10 + cmdReceived[digit_count] - 48;
    digit_count++;
    if (cmdReceived[digit_count] < 48) {
      digit_count = 5;
    }
  }
}

void setTempValue() {
  temp_value = inRange(entered_value, 0, 32766);
  input_mode = 0;
  display_menu = 1;
}

void consoleRefresh() {
  Serial.println();
  Serial.println();
  switch (input_mode) {
    case 0:
      consolePage_1();
      break;
    case 1:
      consolePage_2();
      break;
    case 2:
      analogPinMode();
      break;
    case 3:
      digitalPinMode();
      break;
    case 4:
      valueEntryMode(input_mode);
      break;
    case 5:
      menuJumpMode();
      break;
    case 6:
      valueEntryMode(input_mode);
      break;
    case 7:
      valueEntryMode(input_mode);
      break;
    case 8:
      valueEntryMode(input_mode);
      break;
    case 9:
      valueEntryMode(input_mode);
      break;
    case 10:
      consolePage_3();
      break;
    case 11:
      consolePage_4();
      break;
    case 12:
      consolePage_5();
      break;
    case 13:
      consolePage_6();
      break;
  }
}

void consolePage_1() {
  if (menu_number > 7) {
    transmitText(F("Sequence Menu "));
    Serial.print(menu_number - 7);
  } else {
    transmitText(F("Command Menu "));
    Serial.print(menu_number + 1);
  }
  transmitText(F(" ("));
  Serial.print(menu_number);
  Serial.println(F(")"));
  Serial.println();
  for (int line_number = 0; line_number < 14; line_number++) {
    if (line_number < 9) {
      Serial.print(line_number + 1);
    } else {
      switch (line_number) {
        case 9:
          Serial.print(F("0"));
          break;
        case 10:
          Serial.print(F("E"));
          break;
        case 11:
          Serial.print(F("F"));
          break;
        case 12:
          Serial.print(F("G"));
          break;
        case 13:
          Serial.print(F("H"));
          break;
      }
    }
    Serial.print(F(" - "));
    res_locator(menu_number, line_number, 0); // res_type 0 to print the command description
    Serial.println();
  }
  transmitText(F("Other Commands: A D V W X Y Z C S M ? P Q"));
}

void consolePage_2() {
  Serial.println(F("Help Overview"));
  Serial.println();
  Serial.println(F("The command module responds to single-character commands. No CR is ever required."));
  Serial.println(F("Numher entry may be terminated by CR or by entering the '.' character."));
  Serial.println(F("Command '?' displays this command summary. Alpha commands are not case sensitive."));
  Serial.println(F("Enter the following characters for additional help."));
  Serial.println();
  Serial.println(F("Press L for help controlling servo limits."));
  Serial.println(F("Press N for help with number entry commands: 'A', 'D', 'V', 'W', 'X', 'Y', and 'Z'."));
  Serial.println(F("Press O for help with menu command options 1-9, 0, 'E', 'F', 'G', and 'H'."));
  Serial.println(F("Press P for process and menu display commands: 'Q', Ctrl-C, 'M', 'C', 'S', '+' and '-'."));
  input_mode = 0;
}

void consolePage_3() {
  Serial.println(F("Process and Menu Display Command Help Page"));
  Serial.println();
  Serial.println(F("Process control commands 'Q' and 'Ctrl-C' terminate a sequence or looping process."));
  Serial.println(F("The sequence process sends commands at timed intervals according to a script."));
  Serial.println(F("Scripted commands are preprogrammed and set in code and may loop or not as required."));
  Serial.println();
  Serial.println(F("Commands that control the menu display process do not themselves send commands to other nodes."));
  Serial.println(F("Command 'M' displays the current menu."));
  Serial.println(F("The 'C' and 'S' menu types are merely for convenience."));
  Serial.println(F("'C' menu options may initiate sequences and 'S' menu items may initiate simple commands."));
  Serial.println(F("If the current menu is a 'C' menu the 'C' command followed by a digit jumps to the indicated 'C' menu."));
  Serial.println(F("If the current menu is a 'C' menu the 'S' command switches to the last viewed sequence menu."));
  Serial.println(F("If the current menu is an 'S' menu the 'S' command followed by a digit jumps to that 'S' menu."));
  Serial.println(F("If the current menu is an 'S' menu the 'C' command switches to the last viewed command menu."));
  Serial.println(F("If not entering a value the '+' command will advance to the next menu."));
  Serial.println(F("If not entering a value the '-' command will move back to the previous menu."));
  input_mode = 0;
}

void consolePage_4() {
  Serial.println(F("Options Help Page"));
  Serial.println();
  Serial.println(F("Entry of a numeric character can go one of two ways. A value may be set or a menu option executed."));
  Serial.println(F("When not in number entry mode a numeric character executes an option from the current menu."));
  Serial.println(F("The letters 'E', 'F', 'G' and 'H' also execute options from the current menu."));
  Serial.println(F("This gives a total of up to 14 options on each menu."));
  Serial.println(F("Here is what happens when a menu option is selected."));
  Serial.println();
  Serial.println(F("The menu option description is displayed."));
  Serial.println(F("A command sequence is loaded into memory if the option initiates a sequence. If not ..."));
  Serial.println(F("The command is prepared."));
  Serial.println(F("The command is displayed. (this may be toggled on/off with '/')"));
  Serial.println(F("The command is sent."));
  Serial.println(F("If a sequence is in memory these three steps are repeated as the time elapses for each sequential step."));
  Serial.println();
  Serial.println(F("Replies from the remote node are displayed line by line as they are received."));
  Serial.println(F("Commands that change values in the target module do not (typically) cause a reply to be returned."));
  Serial.println(F("If confirmation of the desired change is needed a second option needs to be selected."));
  Serial.println(F("When a sequence is terminated the command sequence in memory is deleted."));
  Serial.println(F("If additional sequences are needed the Arduino sketch can be modified as required."));
  input_mode = 0;
}

void consolePage_5() {
  Serial.println(F("Number Entry Help Page"));
  Serial.println();
  Serial.println(F("Numeric characters are not treated as commands when entering values but are temporarily saved."));
  Serial.println(F("Commands 'A', 'D', 'V', 'W', 'X', 'Y', and 'Z' set values."));
  Serial.println(F("This continues until the number entry process is terminated."));
  Serial.println(F("If terminated by the '+' character the current pin or value is incremented."));
  Serial.println(F("If terminated by the '-' character the current pin or value is decremented."));
  Serial.println(F("If terminated by the . character or CR the number is stored in the variable."));
  Serial.println(F("Other commands can terminate the process and the variable will stay as it was."));
  Serial.println();
  Serial.println(F("These are the current values of the variables."));
  Serial.print(F("A = "));
  Serial.print(analog_pin);
  Serial.println(F(" (the current Analog Pin)"));
  Serial.print(F("D = "));
  Serial.print(digital_pin);
  Serial.print(F(","));
  prt_digit_pin();
  Serial.println();
  input_mode = 0;
}

void consolePage_6() {
  Serial.println(F("Servo Limit Help Page"));
  Serial.println();
  Serial.println(F("Servo limits are values in the servo module that can limit a servo's range of motion."));
  Serial.println(F("Special commands help control and manipulate these values."));
  Serial.println(F("In the demo code values for servos on digital pins 2, 3, 4 and 5 are manipulated."));
  Serial.println(F("These are the current values of the variables."));
  Serial.println();
  limitReport();
  input_mode = 0;
}

void limitReport() {
  Serial.print(F("Servo on D2: Min = "));
  Serial.print(F(", Max = "));
  Serial.print(F("Servo on D3: Min = "));
  Serial.print(F(", Max = "));
  Serial.print(F("Servo on D4: Min = "));
  Serial.print(F(", Max = "));
  Serial.print(F("Servo on D5: Min = "));
  Serial.print(F(", Max = "));
}

void analogPinMode() {
  if (digit_count < 1) {
    Serial.print(F("Current analog Pin: "));
    Serial.println(analog_pin);
    Serial.print(F("Enter an analog pin #"));
    followedBy();
  } else {
    prtEnteredValue();
  }
}

void digitalPinMode() {
  if (digit_count < 1) {
    Serial.print(F("Current digital pin: "));
    Serial.println(digital_pin);
    Serial.print(F("Enter a digital pin #"));
    followedBy();
  } else {
    prtEnteredValue();
  }
}

void menuJumpMode() {
  if (digit_count < 1) {
    Serial.println(F("Enter a menu # or letter."));
  }
}

void valueEntryMode(int which_value) {
  if (digit_count < 1) {
    Serial.print(F("Current value: "));
    Serial.print(F("Enter a value"));
    followedBy();
  } else {
    prtEnteredValue();
  }
}

void followedBy() {
  Serial.print(F(" followed by . (period) or return."));
}

void prtEnteredValue() {
  Serial.println();
  Serial.print(F("Entered value: "));
  Serial.println(entered_value);
}

// mnu_num is menu number (zero based)
// lne_num is line number (zero based)
// res_type is 0 for command description only, 1 for describe and execute
void res_locator(int mnu_num, int lne_num, int res_type) {
  if (res_type > 0) {
    Serial.println();
    Serial.println();
  }
  switch (mnu_num) {
    case 0:
      menu_num_0_cmd_menu_1(lne_num, res_type);
      break;
    case 1:
      menu_num_1_cmd_menu_2(lne_num, res_type);
      break;
    case 2:
      menu_num_2_cmd_menu_3(lne_num, res_type);
      break;
    case 3:
      menu_num_3_cmd_menu_4(lne_num, res_type);
      break;
    case 4:
      menu_num_4_cmd_menu_5(lne_num, res_type);
      break;
    case 5:
      menu_num_5_cmd_menu_6(lne_num, res_type);
      break;
    case 6:
      menu_num_6_cmd_menu_7(lne_num, res_type);
      break;
    case 7:
      menu_num_7_cmd_menu_8_history(lne_num, res_type);
      break;
    case 8:
      menu_num_8_seq_menu_1(lne_num, res_type);
      break;
    case 9:
      menu_num_9_seq_menu_2(lne_num, res_type);
      break;
  }
}

void menu_num_0_cmd_menu_1(int line_no, int cmd_type) {
  switch (line_no) {
    case 0:
      transmitText(F("S1025 - Set all the digital IO to Input"));
      if (cmd_type > 0) {
        transmitText(F("S1025"));
      }
      break;
    case 1:
      transmitText(F("S1026 - Set the servo module to DEMO mode"));
      if (cmd_type > 0) {
        transmitText(F("S1026"));
      }
      break;
    case 2:
      transmitText(F("S1999 - Comprehensive Status Report"));
      if (cmd_type > 0) {
        transmitText(F("S1999"));
      }
      break;
    case 3:
      transmitText(F("S1998 - All Three Digital Reports"));
      if (cmd_type > 0) {
        transmitText(F("S1998"));
      }
      break;
    case 4:
      transmitText(F("S1997 - Analog Report - Pin Settings"));
      if (cmd_type > 0) {
        transmitText(F("S1997"));
      }
      break;
    case 5:
      transmitText(F("S1996 - Digital Report 1 - Pin Function Settings"));
      if (cmd_type > 0) {
        transmitText(F("S1996"));
      }
      break;
    case 6:
      transmitText(F("S1995 - Digital Report 2 - State, On/Off Times, Var 1/Var 2"));
      if (cmd_type > 0) {
        transmitText(F("S1995"));
      }
      break;
    case 7:
      transmitText(F("S1994 - Digital Report 3 - PWM Settings"));
      if (cmd_type > 0) {
        transmitText(F("S1994"));
      }
      break;
    case 8:
      temp_value = inRange(parm_value_V, 5, 4000);
      transmitText(F("S1024,xxxxx - Set the pulse loop size to "));
      Serial.print(temp_value);
      transmitText(F(" (V: 5-4000)"));
      if (cmd_type > 0) {
        tmpString = F("S1024,");
        tmpString += fiveDigitParm(temp_value);
        transmitText(tmpString);
      }
      break;
    case 9:
      transmitText(F("S1965 - Retrieve the pulse loop size (V)"));
      if (cmd_type > 0) {
        transmitText(F("S1965"));
      }
      break;
    case 10:
      transmitText(F("Synchronize Command Module with Remote Node"));
      if (cmd_type > 0) {
        step_no = 0;
        set_step(100, F("S1965")); // get pulse loop size
        set_step(100, F("S1990")); // get current digital pin
        set_step(100, F("M0305")); // get servo limits
      }
      break;
    case 11: // F
      transmitText(F("S1001 - Set mode to LOCAL"));
      if (cmd_type > 0) {
        transmitText(F("S1001"));
      }
      break;
    case 12: // G
      transmitText(F("S1002 - Set mode to REMOTE"));
      if (cmd_type > 0) {
        transmitText(F("S1002"));
      }
      break;
    case 13: // H
      transmitText(F("S1991 - Retrieve 'LOCAL' or 'REMOTE'"));
      if (cmd_type > 0) {
        transmitText(F("S1991"));
      }
      break;
    default:
      no_command_assigned(0, line_no, cmd_type);
      break;
  }
}

void menu_num_1_cmd_menu_2(int line_no, int cmd_type) {
  switch (line_no) {
    case 0:
      transmitText(F("S1993 - Get settings for"));
      prt_digit_pin();
      if (cmd_type > 0) {
        transmitText(F("S1993"));
      }
      break;
    case 1:
      transmitText(F("S1990 - Retrieve current digital pin #"));
      if (cmd_type > 0) {
        transmitText(F("S1990"));
      }
      break;
    case 2:
      transmitText(F("S1983 - Retrieve 0 or 1 for"));
      prt_digit_pin();
      if (cmd_type > 0) {
        transmitText(F("S1983"));
      }
      break;
    case 3:
      temp_value = inRange(parm_value_V, 0, 4);
      transmitText(F("S1007,"));
      transmitText(twoDigitParm(temp_value));
      transmitText(F(" - Set"));
      prt_digit_pin();
      transmitText(F(" function to "));
      printFunction(temp_value);
      transmitText(F(" (V: 0-4)"));
      if (cmd_type > 0) {
        tmpString = F("S1007,");
        tmpString += twoDigitParm(temp_value);
        transmitText(tmpString);
      }
      break;
    case 4:
      transmitText(F("S1987 - Retrieve 'RAW', 'DEB', 'LED' or 'PWM'"));
      if (cmd_type > 0) {
        transmitText(F("S1987"));
      }
      break;
    case 5:
      transmitText(F("S1005 - Lock"));
      prt_digit_pin();
      if (cmd_type > 0) {
        transmitText(F("S1005"));
      }
      break;
    case 6:
      transmitText(F("S1006 - Unlock"));
      prt_digit_pin();
      if (cmd_type > 0) {
        transmitText(F("S1006"));
      }
      break;
    case 7:
      transmitText(F("S1988 - Retrieve 'SWI' or 'LOC'"));
      if (cmd_type > 0) {
        transmitText(F("S1988"));
      }
      break;
    case 8:
      transmitText(F("S1986 - Retrieve 'IN' or 'OUT'"));
      if (cmd_type > 0) {
        transmitText(F("S1986"));
      }
      break;
    case 9:
      temp_value = inRange(parm_value_W, 0, 5);
      transmitText(F("S1008,"));
      transmitText(twoDigitParm(temp_value));
      transmitText(F(" - Set"));
      prt_digit_pin();
      transmitText(F(" logic to "));
      printLogic(temp_value);
      transmitText(F(" (W: 0-5)"));
      if (cmd_type > 0) {
        tmpString = F("S1008,");
        tmpString += twoDigitParm(temp_value);
        transmitText(tmpString);
      }
      break;
    case 10:
      transmitText(F("S1985 - Retrieve 'REM', 'BTN', 'POT', 'PMN' or 'PTR'"));
      if (cmd_type > 0) {
        transmitText(F("S1985"));
      }
      break;
    case 11:
      temp_value = inRange(parm_value_X, 0, 13);
      transmitText(F("S1009,"));
      transmitText(twoDigitParm(temp_value));
      transmitText(F(" - Set the control input pin for"));
      prt_digit_pin();
      transmitText(F("to "));
      Serial.print(temp_value);
      transmitText(F(" (X: 0-13)"));
      if (cmd_type > 0) {
        tmpString = F("S1009,");
        tmpString += twoDigitParm(temp_value);
        transmitText(tmpString);
      }
      break;
    case 12:
      transmitText(F("S1984 - Retrieve PARM source pin for"));
      prt_digit_pin();
      if (cmd_type > 0) {
        transmitText(F("S1984"));
      }
      break;
    default:
      no_command_assigned(1, line_no, cmd_type);
      break;
  }
}

void menu_num_2_cmd_menu_3(int line_no, int cmd_type) {
  switch (line_no) {
    case 0:
      transmitText(F("S1993 - Get settings for"));
      prt_digit_pin();
      if (cmd_type > 0) {
        transmitText(F("S1993"));
      }
      break;
    case 1:
      transmitText(F("S1015 - Set the direction to REVERSE"));
      if (cmd_type > 0) {
        transmitText(F("S1015"));
      }
      break;
    case 2:
      transmitText(F("S1016 - Set the direction to NORMAL"));
      if (cmd_type > 0) {
        transmitText(F("S1016"));
      }
      break;
    case 3:
      transmitText(F("S1974 - Retrieve 'NORMAL' or 'REVERSED'"));
      if (cmd_type > 0) {
        transmitText(F("S1974"));
      }
      break;
    case 4:
      transmitText(F("S1014,xxxxx - Set the servo slew rate (W)"));
      if (cmd_type > 0) {
        tmpString = F("S1014,");
        tmpString += fiveDigitParm(temp_value);
        transmitText(tmpString);
      }
      break;
    case 5:
      transmitText(F("S1975 - Retrieve the current slew rate"));
      if (cmd_type > 0) {
        transmitText(F("S1975"));
      }
      break;
    case 6:
      transmitText(F("S1980 - Retrieve the current servo pulse width"));
      if (cmd_type > 0) {
        transmitText(F("S1980"));
      }
      break;
    case 7:
      temp_value = inRange(parm_value_X, 100, 3000);
      transmitText(F("S1011,xxxxx - Set the servo target (X)"));
      if (cmd_type > 0) {
        tmpString = F("S1011,");
        tmpString += fiveDigitParm(temp_value);
        transmitText(tmpString);
      }
      break;
    case 8:
      transmitText(F("S1979 - Retrieve the current target pulse width"));
      if (cmd_type > 0) {
        transmitText(F("S1979"));
      }
      break;
    case 13:
      transmitText(F("S1976 - Retrieve the current pulse width range"));
      if (cmd_type > 0) {
        transmitText(F("S1976"));
      }
      break;
    default:
      no_command_assigned(4, line_no, cmd_type);
      break;
  }
}

void menu_num_3_cmd_menu_4(int line_no, int cmd_type) {
  switch (line_no) {
    case 0:
      transmitText(F("S1993 - Get settings for"));
      prt_digit_pin();
      if (cmd_type > 0) {
        transmitText(F("S1993"));
      }
      break;
    case 1:
      transmitText(F("S1003,02 - Set pincer as target"));
      if (cmd_type > 0) {
        digital_pin = 2;
        set_pin_digital();
      }
      break;
    case 2:
      transmitText(F("S1003,03 - Set left/right as target"));
      if (cmd_type > 0) {
        digital_pin = 3;
        set_pin_digital();
      }
      break;
    case 3:
      transmitText(F("S1003,04 - Set up/down as target"));
      if (cmd_type > 0) {
        digital_pin = 4;
        set_pin_digital();
      }
      break;
    case 4:
      transmitText(F("S1003,05 - Set out/in as target"));
      if (cmd_type > 0) {
        digital_pin = 5;
        set_pin_digital();
      }
      break;
    case 5:
      transmitText(F("Get Min and Max for demo servos"));
      if (cmd_type > 0) {
        step_no = 0;
        set_step(100, F(">"));     // save current digital pin
        set_step(100, F("M0304")); // set out/in as target
        set_step(100, F("S1978")); // get min pulse width
        set_step(100, F("S1977")); // get max pulse width
        set_step(100, F("M0303")); // set up/down as target
        set_step(100, F("S1978")); // get min pulse width
        set_step(100, F("S1977")); // get max pulse width
        set_step(100, F("M0302")); // set left/right as target
        set_step(100, F("S1978")); // get min pulse width
        set_step(100, F("S1977")); // get max pulse width
        set_step(100, F("M0301")); // set pincer as target
        set_step(100, F("S1978")); // get min pulse width
        set_step(100, F("S1977")); // get max pulse width
        set_step(100, F("<"));     // restore current digital pin
      }
      break;
    case 6:
      temp_value = inRange(parm_value_Y, 100, 3000);
      transmitText(F("S1012,xxxxx - Set the min servo pulse width (Y)"));
      if (cmd_type > 0) {
        tmpString = F("S1012,");
        tmpString += fiveDigitParm(temp_value);
        transmitText(tmpString);
      }
      break;
    case 7:
      transmitText(F("S1978 - Retrieve the current min pulse width (Y)"));
      if (cmd_type > 0) {
        transmitText(F("S1978"));
      }
      break;
    case 8:
      temp_value = inRange(parm_value_Z, 300, 3000);
      transmitText(F("S1013,xxxxx - Set the max servo pulse width (Z)"));
      if (cmd_type > 0) {
        tmpString = F("S1013,");
        tmpString += fiveDigitParm(temp_value);
        transmitText(tmpString);
      }
      break;
    case 9:
      transmitText(F("S1977 - Retrieve the current max pulse width (Z)"));
      if (cmd_type > 0) {
        transmitText(F("S1977"));
      }
      break;
    case 10:
      transmitText(F("Move Servo cconnected to D"));
      Serial.print(digital_pin);
      if (cmd_type > 0) {
        step_no = 0;
        set_step(100, F("S1008,00"));  // use REMOTE control
        set_step(100, F("S1011,01200")); // set servo target
        set_step(3999, F("S1011,02000")); // set servo target
        set_step(3999, F("S1011,01200")); // set servo target
        set_step(3899, F("R0205")); // return 2 steps
      }
      break;
    case 11:
      transmitText(F("Motion Combo #1"));
      if (cmd_type > 0) {
        step_no = 0;
        set_step(100, F("M0304")); // set out/in as target
        set_step(100, F("S1008,00"));  // use REMOTE control
        set_step(100, F("S1011,01200")); // set servo target
        set_step(100, F("M0303")); // set up/down as target
        set_step(100, F("S1008,00"));  // use REMOTE control
        set_step(100, F("S1011,01200")); // set servo target
        set_step(3999, F("M0304")); // set out/in as target
        set_step(100, F("S1011,02000")); // set servo target
        set_step(100, F("M0303")); // set up/down as target
        set_step(100, F("S1011,02000")); // set servo target
        set_step(3899, F("R1005")); // return 10 steps
      }
      break;
    case 12:
      transmitText(F("Motion Combo #2"));
      if (cmd_type > 0) {
        step_no = 0;
        set_step(100, F("M0302")); // set left/right as target
        set_step(100, F("S1008,00"));  // use REMOTE control
        set_step(100, F("S1011,01200")); // set servo target
        set_step(100, F("M0303")); // set up/down as target
        set_step(100, F("S1008,00"));  // use REMOTE control
        set_step(100, F("S1011,01200")); // set servo target
        set_step(3999, F("M0302")); // set left/right as target
        set_step(100, F("S1011,02000")); // set servo target
        set_step(100, F("M0303")); // set up/down as target
        set_step(100, F("S1011,02000")); // set servo target
        set_step(3899, F("R1005")); // return 10 steps
      }
      break;
    default:
      no_command_assigned(2, line_no, cmd_type);
      break;
  }
}

void menu_num_4_cmd_menu_5(int line_no, int cmd_type) {
  switch (line_no) {
    default:
      no_command_assigned(3, line_no, cmd_type);
      break;
  }
}

void menu_num_5_cmd_menu_6(int line_no, int cmd_type) {
  switch (line_no) {
    case 0:
      transmitText(F("S1993 - Get settings for"));
      prt_digit_pin();
      if (cmd_type > 0) {
        transmitText(F("S1993"));
      }
      break;
    case 1:
      transmitText(F("S1017 - Set the trigger on"));
      prt_digit_pin();
      if (cmd_type > 0) {
        transmitText(F("S1017"));
      }
      break;
    case 2:
      transmitText(F("S1973 - Retrieve the current toggle state"));
      if (cmd_type > 0) {
        transmitText(F("S1973"));
      }
      break;
    case 3:
      transmitText(F("S1027 - Turn tracking on"));
      if (cmd_type > 0) {
        transmitText(F("S1027"));
      }
      break;
    case 4:
      transmitText(F("S1028 - Turn tracking off"));
      if (cmd_type > 0) {
        transmitText(F("S1028"));
      }
      break;
    case 5:
      transmitText(F("S1964 - Return tracker state"));
      if (cmd_type > 0) {
        transmitText(F("S1964"));
      }
      break;
    case 6:
      temp_value = inRange(parm_value_Y, 100, 3000);
      transmitText(F("S1010,xxxxx - Set the LED brightness for"));
      prt_digit_pin();
      transmitText(F(" (X)"));
      if (cmd_type > 0) {
        tmpString = F("S1010,");
        tmpString += fiveDigitParm(temp_value);
        transmitText(tmpString);
      }
      break;
    case 7:
      transmitText(F("S1982 - Retrieve on-time for"));
      prt_digit_pin();
      if (cmd_type > 0) {
        transmitText(F("S1982"));
      }
      break;
    case 8:
      transmitText(F("S1981 - Retrieve off-time for"));
      prt_digit_pin();
      if (cmd_type > 0) {
        transmitText(F("S1981"));
      }
      break;
    case 9:
      temp_value = inRange(parm_value_Y, 100, 3000);
      transmitText(F("S1018,xxxxx - Set digital var #1 (Y)"));
      if (cmd_type > 0) {
        tmpString = F("S1018,");
        tmpString += fiveDigitParm(temp_value);
        transmitText(tmpString);
      }
      break;
    case 10:
      transmitText(F("S1972 - Retrieve digital var #1"));
      if (cmd_type > 0) {
        transmitText(F("S1972"));
      }
      break;
    case 11:
      temp_value = inRange(parm_value_Y, 100, 3000);
      transmitText(F("S1019,xxxxx - Set digital var #2 (Z)"));
      if (cmd_type > 0) {
        tmpString = F("S1019,");
        tmpString += fiveDigitParm(temp_value);
        transmitText(tmpString);
      }
      break;
    case 12:
      transmitText(F("S1971 - Retrieve digital var #2"));
      if (cmd_type > 0) {
        transmitText(F("S1971"));
      }
      break;
    default:
      no_command_assigned(5, line_no, cmd_type);
      break;
  }
}

void menu_num_6_cmd_menu_7(int line_no, int cmd_type) {
  switch (line_no) {
    case 0:
      transmitText(F("S1992 - Current Analog Pin Settings"));
      if (cmd_type > 0) {
        transmitText(F("S1992"));
      }
      break;
    case 1:
      transmitText(F("S1989 - Retrieve the current analog pin #"));
      if (cmd_type > 0) {
        transmitText(F("S1989"));
      }
      break;
    case 2:
      transmitText(F("S1970 - Retrieve the analog input value"));
      if (cmd_type > 0) {
        transmitText(F("S1970"));
      }
      break;
    case 3:
      temp_value = inRange(parm_value_W, 300, 3000);
      transmitText(F("S1020,xxxxx - Set the analog lower bound (W)"));
      if (cmd_type > 0) {
        tmpString = F("S1020,");
        tmpString += fiveDigitParm(temp_value);
        transmitText(tmpString);
      }
      break;
    case 4:
      transmitText(F("S1969 - Retrieve the lower bound"));
      if (cmd_type > 0) {
        transmitText(F("S1969"));
      }
      break;
    case 5:
      temp_value = inRange(parm_value_X, 300, 3000);
      transmitText(F("S1021,xxxxx - Set the analog upper bound (X)"));
      if (cmd_type > 0) {
        tmpString = F("S1021,");
        tmpString += fiveDigitParm(temp_value);
        transmitText(tmpString);
      }
      break;
    case 6:
      transmitText(F("S1968 - Retrieve the upper bound"));
      if (cmd_type > 0) {
        transmitText(F("S1968"));
      }
      break;
    case 7:
      transmitText(F("S1967 - Retrieve the range"));
      if (cmd_type > 0) {
        transmitText(F("S1967"));
      }
      break;
    case 8:
      transmitText(F("S1966 - Retrieve the scaled input value"));
      if (cmd_type > 0) {
        transmitText(F("S1966"));
      }
      break;
    case 9:
      temp_value = inRange(parm_value_Y, 300, 3000);
      transmitText(F("S1022,xxxxx - Set analog var #1 (Y)"));
      if (cmd_type > 0) {
        tmpString = F("S1022,");
        tmpString += fiveDigitParm(temp_value);
        transmitText(tmpString);
      }
      break;
    case 10:
      transmitText(F("S1963 - Retrieve analog var #1"));
      if (cmd_type > 0) {
        transmitText(F("S1963"));
      }
      break;
    case 11:
      temp_value = inRange(parm_value_Z, 300, 3000);
      transmitText(F("S1023,xxxxx - Set analog var #2 (Z)"));
      if (cmd_type > 0) {
        tmpString = F("S1023,");
        tmpString += fiveDigitParm(temp_value);
        transmitText(tmpString);
      }
      break;
    case 12:
      transmitText(F("S1962 - Retrieve analog var #2"));
      if (cmd_type > 0) {
        transmitText(F("S1962"));
      }
      break;
    case 13:
      transmitText(F("S1029 - Toggle analog toggle state"));
      if (cmd_type > 0) {
        transmitText(F("S1029"));
      }
      break;
    default:
      no_command_assigned(6, line_no, cmd_type);
      break;
  }
}

void no_command_assigned(int menu, int key, int cmd_type) {
  if (cmd_type > 0) {
    transmitText(F("Command Menu "));
    Serial.print(menu + 1);
    optionNotDefined(key);
  }
}

void menu_num_7_cmd_menu_8_history(int line_no, int cmd_type) {
  if (strlen(cmdHistory[line_no]) > 0) {
    transmitText(F("Send ("));
    Serial.print(strlen(cmdHistory[line_no]));
    transmitText(F(") "));
    transmitText(cmdHistory[line_no]);
    if (cmd_type > 0) {
      transmitText(cmdHistory[line_no]);
    }
  }
}

void menu_num_8_seq_menu_1(int line_no, int cmd_type) {
  switch (line_no) {
    case 1:
      transmitText(F("Set Pulse Loop Size"));
      if (cmd_type > 0) {
        step_no = 0;
        set_step(200, F("S1024,00005"));
        set_step(200, F("S1024,00010"));
        set_step(200, F("S1024,00050"));
        set_step(200, F("S1024,00100"));
        set_step(200, F("S1024,00150"));
        set_step(200, F("S1024,00025"));
        set_step(200, F("S1965"));
      }
      break;
    case 2:
      transmitText(F("Set Servo slew rate for D"));
      Serial.print(digital_pin);
      if (cmd_type > 0) {
        step_no = 0;
        set_step(200, F("S1975"));
        set_step(100, F("S1014,00000")); // set slew rate
        set_step(100, F("S1014,00001")); // set slew rate
        set_step(100, F("S1014,00002")); // set slew rate
        set_step(100, F("S1014,00004")); // set slew rate
        set_step(100, F("S1014,00008")); // set slew rate
        set_step(100, F("S1014,00012")); // set slew rate
        set_step(200, F("S1975"));
      }
      break;
    case 3:
      transmitText(F("Get settings for"));
      prt_digit_pin();
      if (cmd_type > 0) {
        step_no = 0;
        set_step(100, F("S1990"));
        set_step(100, F("S1987"));
        set_step(100, F("S1988"));
        set_step(100, F("S1986"));
        set_step(100, F("S1985"));
        set_step(100, F("S1984"));
        set_step(100, F("S1980"));
        set_step(100, F("S1979"));
        set_step(100, F("S1975"));
        set_step(100, F("S1028"));
        set_step(100, F("P1"));
      }
      break;
    default:
      if (cmd_type > 0) {
        transmitText(F("Sequence Menu 1"));
        optionNotDefined(line_no);
      }
      break;
  }
}

void menu_num_9_seq_menu_2(int line_no, int cmd_type) {
  switch (line_no) {
    case 0:
      transmitText(F("Blink LED cconnected to D"));
      Serial.print(digital_pin);
      if (cmd_type > 0) {
        step_no = 0;
        set_step(100, F("S1006"));     // unlock pin
        set_step(100, F("S1007,02"));  // make it an LED output
        set_step(100, F("S1008,00"));  // use REMOTE control
        set_step(100, F("S1010,00100")); // turn LED on
        set_step(999, F("S1010,00000")); // turn LED off
        set_step(999, F("S1010,00100")); // turn LED on
        set_step(10, F("R0300")); // return 3 steps
      }
      break;
    case 6:
      transmitText(F("S1003,09 - Set LED 1 as target"));
      if (cmd_type > 0) {
        digital_pin = 9;
        set_pin_digital();
        display_menu = 1;
      }
      break;
    case 7:
      transmitText(F("S1003,11 - Set LED 2 as target"));
      if (cmd_type > 0) {
        digital_pin = 11;
        set_pin_digital();
        display_menu = 1;
      }
      break;
    case 8:
      transmitText(F("S1003,13 - Set LED 3 as target"));
      if (cmd_type > 0) {
        digital_pin = 13;
        set_pin_digital();
        display_menu = 1;
      }
      break;
    case 9:
      transmitText(F("S1003,10 - Set LED 4 as target"));
      if (cmd_type > 0) {
        digital_pin = 10;
        set_pin_digital();
        display_menu = 1;
      }
      break;
    case 10:
      transmitText(F("S1003,8 - Set LED 5 as target"));
      if (cmd_type > 0) {
        digital_pin = 8;
        set_pin_digital();
        display_menu = 1;
      }
      break;
    case 11:
      transmitText(F("S1003,12 - Set LED 6 as target"));
      if (cmd_type > 0) {
        digital_pin = 12;
        set_pin_digital();
        display_menu = 1;
      }
      break;
    case 12:
      transmitText(F("S1003,07 - Set LED 7 as target"));
      if (cmd_type > 0) {
        digital_pin = 7;
        set_pin_digital();
        display_menu = 1;
      }
      break;
    default:
      if (cmd_type > 0) {
        transmitText(F("Sequence Menu 2"));
        optionNotDefined(line_no);
      }
      break;
  }
}

void set_pin_digital() {
  temp_value = inRange(digital_pin, 2, 13);
  tmpString = F("S1003,");
  tmpString += twoDigitParm(temp_value);
  transmitText(tmpString);
}

void prt_digit_pin() {
  transmitText(F(" digital pin D"));
  Serial.print(digital_pin);
}

void optionNotDefined(int option_number) {
  transmitText(F(" Option "));
  option_number++;
  switch (option_number) {
    case 10:
      transmitText(F("0"));
      break;
    case 11:
      transmitText(F("E"));
      break;
    case 12:
      transmitText(F("F"));
      break;
    case 13:
      transmitText(F("G"));
      break;
    case 14:
      transmitText(F("H"));
      break;
    default:
      Serial.print(option_number);
      break;
  }
  transmitText(F(" was selected but not defined"));
}

void printFunction(int fnc_number) {
  switch (fnc_number) {
    case 0:
      transmitText(F("RAW"));
      break;
    case 1:
      transmitText(F("DEB"));
      break;
    case 2:
      transmitText(F("LED"));
      break;
    case 3:
      transmitText(F("PWM"));
      break;
    case 4:
      transmitText(F("PWA"));
      break;
  }
}

void printLogic(int logic_number) {
  switch (logic_number) {
    case 0:
      transmitText(F("REM"));
      break;
    case 1:
      transmitText(F("BTN"));
      break;
    case 2:
      transmitText(F("POT"));
      break;
    case 3:
      transmitText(F("PMN"));
      break;
    case 4:
      transmitText(F("PTR"));
      break;
    case 5:
      transmitText(F("PTA"));
      break;
  }
}

void setCurrentAnalogPin() {
  temp_value = inRange(analog_pin, 0, 7);
  tmpString = F("S1004,");
  tmpString += twoDigitParm(temp_value);
  transmitText(tmpString);
}

// this function merely stores the steps in an array
void set_step(int cmd_delay, const String & cmd_str) {
  if (step_no < NUMBER_OF_ELEMENTS) {
    cmd_str.toCharArray(seqCommand[step_no], 12);
    seqDelay[step_no] = cmd_delay;
    step_no++;
    sequence_steps_to_do = step_no;
    seq_step = 0;
  } else {
    Serial.println(F("It seems there are too many steps defined for this sequence"));
  }
}

void report_one() {
  Serial.println(F("Report 0ne"));
  Serial.println();
  Serial.print(F("Pin D"));
  Serial.print(digital_pin);
  Serial.print(F(", Function: "));
  Serial.print(savResponse[2]);
  Serial.print(savResponse[3]);
  Serial.print(savResponse[4]);
  Serial.print(F(", SWI/LOC: "));
  Serial.print(savResponse[5]);
  Serial.print(savResponse[6]);
  Serial.print(savResponse[7]);
  Serial.print(F(", Direction: "));
  Serial.print(savResponse[8]);
  Serial.print(savResponse[9]);
  Serial.println(savResponse[10]);
  Serial.print(F("Control: "));
  Serial.print(savResponse[11]);
  Serial.print(savResponse[12]);
  Serial.print(savResponse[13]);
  Serial.print(savResponse[14]);
  Serial.print(savResponse[15]);
  Serial.print(savResponse[16]);
  Serial.print(savResponse[17]);
  Serial.print(savResponse[18]);
  Serial.print(F(", Source Pin: "));
  Serial.println(sourcePin);
  Serial.print(F("Current Pulse Width: "));
  Serial.print(pulseWidth);
  Serial.print(F(", Target Pulse Width: "));
  Serial.println(parm_value_X);
  Serial.print(F("Slew Rate "));
  Serial.println(parm_value_W);
}

void report_two() {
  Serial.println(F("Report Two"));
  Serial.println();
  limitReport();
  Serial.println();
}

void transmitText(String message_text) {
  myStr = "";
  int isDuplicate = 0;
  for (h_idx = 0; h_idx < 14; h_idx++) {
    myStr = "";
    myStr.concat(cmdHistory[h_idx]);
    if (message_text.equals(myStr)) {
      isDuplicate = 1;
      h_idx = 14;
    }
  }
  if (isDuplicate < 1) {
    for (h_idx = 0; h_idx < 13; h_idx++) {
      myStr = "";
      myStr.concat(cmdHistory[h_idx + 1]);
      myStr.toCharArray(cmdHistory[h_idx], 12);
    }
    message_text.toCharArray(cmdHistory[13], 12);
  }
  Serial.println();
  Serial.println(message_text);
  for (int i = 0; i < message_text.length(); i++) {
    gtSerial.print(message_text.charAt(i));
  }
  // the first character of the message contains the message length
  // it always follows a period during which the output is HIGH
  // this only needs to be long enough to print the message
  Serial.print(message_text.length());
  Serial.println(F(" characters"));
  delayMicroseconds(START_MSG_PERIOD);
  transmitChar(message_text.length());
  for (int i = 0; i < message_text.length(); i++) {
    transmitChar(message_text.charAt(i));
  }
}

void transmitChar(int value) {
  // seven bits make an ASCII character
  // the bits of the character are always followed by a period during which the output is HIGH
  // this is to allow the receiver to store the character in the message array
  // the text is transmitted without pause other than for that slight delay
  // for consistent timing the pulse durations are in eight simple variables
  // the receiving node is completely dedicated to saving the message chars
  // short period means bit is a zero
  // long period means bit is a one
  // 1 2 3 4 5 6 7 END
  // 3+3+3+3+3+3+3+30   shortest character 0x00 51 microseconds
  // 8+8+8+8+8+8+8+30   longest character  0x7F 86 microseconds
  // min throughput of 11627 characters per second
  // max throughput of 19607 characters per second
  if (value > 127) {
    value = 127;
  }
  if (value > 63) {
    value -= 64;
    tx_period_1 = LONG_PERIOD;
  } else {
    tx_period_1 = SHORT_PERIOD;
  }
  if (value > 31) {
    value -= 32;
    tx_period_2 = LONG_PERIOD;
  } else {
    tx_period_2 = SHORT_PERIOD;
  }
  if (value > 15) {
    value -= 16;
    tx_period_3 = LONG_PERIOD;
  } else {
    tx_period_3 = SHORT_PERIOD;
  }
  if (value > 7) {
    value -= 8;
    tx_period_4 = LONG_PERIOD;
  } else {
    tx_period_4 = SHORT_PERIOD;
  }
  if (value > 3) {
    value -= 4;
    tx_period_5 = LONG_PERIOD;
  } else {
    tx_period_5 = SHORT_PERIOD;
  }
  if (value > 1) {
    value -= 2;
    tx_period_6 = LONG_PERIOD;
  } else {
    tx_period_6 = SHORT_PERIOD;
  }
  if (value > 0) {
    tx_period_7 = LONG_PERIOD;
  } else {
    tx_period_7 = SHORT_PERIOD;
  }
  digitalWrite(13, LOW);
  delayMicroseconds(tx_period_1);
  digitalWrite(13, HIGH);
  delayMicroseconds(tx_period_2);
  digitalWrite(13, LOW);
  delayMicroseconds(tx_period_3);
  digitalWrite(13, HIGH);
  delayMicroseconds(tx_period_4);
  digitalWrite(13, LOW);
  delayMicroseconds(tx_period_5);
  digitalWrite(13, HIGH);
  delayMicroseconds(tx_period_6);
  digitalWrite(13, LOW);
  delayMicroseconds(tx_period_7);
  digitalWrite(13, HIGH);
  delayMicroseconds(END_CHAR_PERIOD);
}

String fiveDigitParm(int value) {
  String retString = "";
  if (value < 0) {
    value = 0;
  }
  if (value < 10000) {
    retString += F("0");
  }
  if (value < 1000) {
    retString += F("0");
  }
  if (value < 100) {
    retString += F("0");
  }
  if (value < 10) {
    retString += F("0");
  }
  retString += value;
  return retString;
}

int inRange(int value, int min_val, int max_val) {
  if (value < min_val) {
    value = min_val;
  }
  if (value > max_val) {
    value = max_val;
  }
  return value;
}

String twoDigitParm(int value) {
  String retString = "";
  if (value < 10) {
    retString += F("0");
  }
  retString += value;
  return retString;
}
